﻿using System;

namespace ELTE.Basics.NumberTypes
{
    /// <summary>
    /// Racionális szám típusa
    /// </summary>
    class Rational : IRealCompatible // igazából : Object, IRealCompatible
        // megvalósítja az IRealCompatible interfészt
    {
        private Int32 _Nominator;
        private Int32 _Denominator;

        /// <summary>
        /// Számláló lekérdezés vagy beállítása
        /// </summary>
        public Int32 Nominator
        {
            get { return _Nominator; }
            set { _Nominator = value; }
        }
        /// <summary>
        /// Nevező lekérdezés vagy beállítása
        /// </summary>
        public Int32 Denominator
        {
            get { return _Denominator; }
            set { _Denominator = (value != 0) ? value : 1; } // a nevező nem lehet 0
        }

        /// <summary>
        /// Racionális szám létrehozása
        /// </summary>
        public Rational()
        {
            Nominator = 0;
            Denominator = 1;
        }

        /// <summary>
        /// Racionális szám létrehozása
        /// </summary>
        /// <param name="nom">Számláló</param>
        /// <param name="den">Nevező</param>
        public Rational(Int32 nom, Int32 den)
        {
            Nominator = nom;
            Denominator = den;
            Simplify(ref _Nominator, ref _Denominator); // az egyszerűsítés referencia szerint veszi át az értékeket
        }

        /// <summary>
        /// Racionális számok összeadása
        /// </summary>
        /// <param name="first">Első operandus</param>
        /// <param name="second">Második operandus</param>
        /// <returns>Összeadás eredménye</returns>
        public static Rational operator +(Rational first, Rational second)
            // az operátorokat statikus metódusként tudjuk megadni
        {
            Rational result = new Rational(); // eredmény
            result.Nominator = first.Nominator * second.Denominator + first.Denominator * second.Nominator; // megfelelő műveletek elvégzése
            result.Denominator = first.Denominator * second.Denominator;
            Simplify(ref result._Nominator, ref result._Denominator); // egyszerűsítés
            return result; // eredmény visszaadása
        }

        /// <summary>
        /// Racionális számok különbsége
        /// </summary>
        /// <param name="first">Első operandus</param>
        /// <param name="second">Második operandus</param>
        /// <returns>Kivonás eredménye</returns>
        public static Rational operator -(Rational first, Rational second)
        {
            Rational result = new Rational();
            result.Nominator = first.Nominator * second.Denominator - first.Denominator * second.Nominator;
            result.Denominator = first.Denominator * second.Denominator;
            Simplify(ref result._Nominator, ref result._Denominator);
            return result;
        }

        /// <summary>
        /// Racionális számok szorzása
        /// </summary>
        /// <param name="first">Első operandus</param>
        /// <param name="second">Második operandus</param>
        /// <returns>Szorzás eredménye</returns>
        public static Rational operator *(Rational first, Rational second)
        {
            Rational result = new Rational();
            result.Nominator = first.Nominator * second.Nominator;
            result.Denominator = first.Denominator * second.Denominator;
            Simplify(ref result._Nominator, ref result._Denominator);
            return result;
        }

        /// <summary>
        /// Racionális számok hányadosa
        /// </summary>
        /// <param name="first">Első operandus</param>
        /// <param name="second">Második operandus</param>
        /// <returns>Osztás eredménye</returns>
        public static Rational operator /(Rational first, Rational second)
        {
            Rational result = new Rational();
            result.Nominator = first.Nominator * second.Denominator;
            result.Denominator = first.Denominator * second.Nominator;
            Simplify(ref result._Nominator, ref result._Denominator);
            return result;
        }

        /// <summary>
        /// Egyszerűsítés művelete
        /// </summary>
        /// <param name="first">Első érték</param>
        /// <param name="second">Második érték</param>
        static private void Simplify(ref Int32 first, ref Int32 second) // a paraméterket referencia szerint adjuk át
        {
            Int32 m = first, n = second;
            if (m < 1 || n < 1)
                return;
            while (m != n) // Euklédeszi algoritmus futtatása
            {
                if (m > n)
                    m -= n;
                else
                    n -= m;
            }
            first /= m; // egyszerűsítés a legnagyobb közös osztóval
            second /= m;
        }

        /// <summary>
        /// Kiírás szövegként
        /// </summary>
        /// <returns>Racionális szám szövegszerű leírása</returns>
        public override String ToString() // felüldefiniáljuk az object alapértelmezett kiírását
        {
            return Nominator.ToString() + "/" + Denominator.ToString();
        }

        #region IRealCompatible Members

        public Double ToDouble()
        {
            return 1.0 * Nominator / Denominator; // a szorzás a koverzióhoz kell
        }

        public void FromDouble(Double val)
        {
            Nominator = Convert.ToInt32(Math.Round(val, 5, MidpointRounding.ToEven) * 10000); // kerekítés 5 tizedesjegyre
            Denominator = 10000;
            Simplify(ref _Nominator, ref _Denominator);
        }

        #endregion

        /// <summary>
        /// Explicit konverziós operátor.
        /// </summary>
        /// <param name="value">Racionális érték.</param>
        /// <returns>Lebegőpontos valós érték.</returns>
        public static explicit operator Double(Rational value)
        {
            return value.ToDouble();
        }
    }
}
